/****************************************************************************
*
*  Structural analysis and modifications to MPEG audio files
*
*  Create:  28.09.99
*  Update:  21.02.04
*  (C)      Marcel Mller 1999-2004
*
****************************************************************************/


#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <limits.h>
#include <assert.h>
#include <share.h>
#include <io.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

#undef min  // risky macros
#undef max

#include <vector>
#include <list>
#include <memory>
#include <algorithm> // replace the min/max macros with side effects
using namespace std;
#include <MMSmartPointer.h>

//#include <../../MMutil+/MMDebugMem.cpp>


// Error handling ***********************************************************
enum EventClass
{  Warning = 3,
   Error
};

static const char* Class2String(EventClass cls)
{  switch (cls)
   {default:
      return "unknown";
    case Warning:
      return "Warning";
    case Error:
      return "Error";
}  }

struct Event
{  int err;
   char buf[1024];
   Event(int err, const char* fmt, ...);
};

Event::Event(int err, const char* fmt, ...)
 : err(err)
{  va_list va;
   va_start(va, fmt);
   vsprintf(buf, fmt, va);
   va_end(va);
}


// mpeg tables **************************************************************
static const char sVersion[4][4] = {"2.5", {'','0','1',''}, "2", "1"};
static const char sLayer[4][4] = {{'','0','0',''}, "III", "II", "I"};
static const char sYesNo[2][4] = {"no", "yes"};
static const int iBitRate[2][4][16] = {
{  {-2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2},
   {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1},
   {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, -1},
   {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1}
},
{  {-2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2},
   {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1},
   {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, -1},
   {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1}
} };
static const int iSampleRate[4][4] =
{  {11025, 12000, 8000, -1},
   {-2, -2, -2, -2},
   {22050, 24000, 16000, -1},
   {44100, 48000, 32000, -1}
};
static const char sMode[4][2] = {"S", {'J','S'}, {'D','C'}, {'S','C'}};
static const char sExtMode[2][4][8] =
{  {"4-31", "8-31", "12-31", "16-31"},
   {"off off", "on  off", "off on", "on  on"}
};
static const char sSExtMode[2][4][8] =
{  {"IS 4-31", "IS 8-31", {'I','S',' ','1','2','-','3','1'}, {'I','S',' ','1','6','-','3','1'}},
   {"normal", "IS", "MS-S", "IS+MS-S"}
};
static const char sCopyright[4][5] = {"none", "none", {'o','r','i','g','.'}, "copy"};
static const char sEmphasis[4][5] = {"none", {'5','0','/','1','5'}, "10", "CCIT"};

static const unsigned iSideInfoLen[4] = {9, 17, 17, 32};


// mpeg structs (little endian) *********************************************
struct MPEG_Frame_Header
{  // layout
   signed   FrameSync1    : 8;
   unsigned CRC           : 1;
   unsigned Layer         : 2;
   unsigned Version       : 2;
   signed   FrameSync2    : 3;
   unsigned Private       : 1;
   unsigned Padding       : 1;
   unsigned SamplingRate  : 2;
   unsigned BitRate       : 4;
   unsigned Emphasis      : 2;
   unsigned Copyright     : 2;
   unsigned ModeExtension : 2;
   unsigned Mode          : 2;
   // methods
   int      GetLayer() const           { return 4 - Layer; }
   int      GetSamplingRate() const    { return iSampleRate[Version][SamplingRate]; }
   const int* GetBitRates() const      { return iBitRate[Version != 3][Layer]; }
   int      GetBitRate(const int* brv) const { return brv[BitRate]; }
   int      GetBitRate() const         { return iBitRate[Version != 3][Layer][BitRate]; }
   size_t   GetFrameLen(int br) const;
   size_t   GetFrameLen() const        { return GetFrameLen(GetBitRate()); }
   bool     HasCRC() const             { return !CRC; }
   bool     Has2Channels() const       { return Mode != 3 && Layer == 1; }
   size_t   GetFrameSamples() const    { return Layer == 3 ? 384U : 1152U; }
   size_t   GetGranules() const        { return (Version == 3) +1; }
};

size_t MPEG_Frame_Header::GetFrameLen(int br) const
{  int sr = iSampleRate[3][SamplingRate];
   return (br <= 0 || sr <= 0 || Version == 1)
    ? 0 : (Layer == 3)
      ? (12000 * br / sr + Padding) << 2
      : ((144000 * br) << (Version != 3)) / sr + Padding; // I _really_ don't know why the >> operation is needed.
}


// XING VBR header funcs ****************************************************
static int ExtractI4(const unsigned char* &buf)
{  // big endian extract
   int x = (((buf[0] << 8) + buf[1] << 8) + buf[2] << 8) + buf[3];
   buf += 4;
   return x;
}

static void StoreI4(unsigned char* &buf, int i)
{  // big endian store
   buf[3] = (unsigned char)i;
   i >>= 8;
   buf[2] = (unsigned char)i;
   i >>= 8;
   buf[1] = (unsigned char)i;
   i >>= 8;
   buf[0] = (unsigned char)i;
   buf += 4;
}

#define FRAMES_FLAG     0x0001
#define BYTES_FLAG      0x0002
#define TOC_FLAG        0x0004
#define VBR_SCALE_FLAG  0x0008

#define FRAMES_AND_BYTES (FRAMES_FLAG | BYTES_FLAG)
// structure to receive extracted header
// toc may be NULL
struct XHEADDATA
{  int h_id;       // from MPEG header, 3=MPEG1, 2=MPEG2, 0=MPEG2.5
   int h_mode;     // from MPEG header, channel mode, 3=mono
   int samprate;   // determined from MPEG header
   int flags;      // from Xing header data
   int frames;     // total bit stream frames from Xing header data
   int bytes;      // total bit stream bytes from Xing header data
   int vbr_scale;  // encoded vbr scale from Xing header data
   unsigned char toc[100];// toc_buffer[100]

   static auto_ptr<XHEADDATA> Get(const unsigned char *buf);
   void Put(unsigned char *buf) const;
   void Recalculate(size_t _frames, size_t _bytes, size_t* frameloc, size_t offset = 0);
   size_t Offset() const      { return 4 + iSideInfoLen[((h_id == 3) << 1) + (h_mode != 3)]; }
   size_t Length() const;
   #ifdef DEBUG
   void Print() const;
   #endif
};

#ifdef DEBUG
void XHEADDATA::Print() const
{  printf("*** XING-Header ***\n"
          "ID            %i\n"
          "Samplingrate  %i\n"
          "Flags         %X\n"
          "Frames        %i\n"
          "Bytes         %i\n"
          "VBR scale     %i\n",
    h_id, samprate, flags, frames, bytes, vbr_scale);
   for (int i = 0; i < 100; ++i)
      printf("%3i ", toc[i]);
   puts("");
}
#endif

auto_ptr<XHEADDATA> XHEADDATA::Get(const unsigned char *buf)
// return 0=fail, 1=success
// X   structure to receive header data (output)
// buf bitstream input
{  auto_ptr<XHEADDATA> X(new XHEADDATA);

   // get Xing header data
   X->flags = 0;     // clear to null incase fail

   // get selected MPEG header data
   X->h_id     = (buf[1] >> 3) & 3;
   X->samprate = iSampleRate[X->h_id][(buf[2] >> 2) & 3];
   X->h_mode   = (buf[3] >> 6) & 3;

   // determine offset of header
   buf += X->Offset();

   if (buf[0] != 'X' || buf[1] != 'i' || buf[2] != 'n' || buf[3] != 'g')
      return auto_ptr<XHEADDATA>();              // fail, header not found
   buf+=4;

   X->flags = ExtractI4(buf);// get flags

   if (X->flags & FRAMES_FLAG)
      X->frames   = ExtractI4(buf);
   if (X->flags & BYTES_FLAG)
      X->bytes = ExtractI4(buf);

   if (X->flags & TOC_FLAG)
      memcpy(X->toc, buf, 100), buf+=100;

   X->vbr_scale = -1;
   if (X->flags & VBR_SCALE_FLAG)
      X->vbr_scale = ExtractI4(buf);

   #ifdef DEBUG
   X->Print();
   #endif

   return X;       // success
}

void XHEADDATA::Put(unsigned char *buf) const
// X   structure to header data (input)
// buf bitstream input/output
{  // determine offset of header
   buf += Offset();

   buf[0] = 'X';
   buf[1] = 'i';
   buf[2] = 'n';
   buf[3] = 'g';
   buf+=4;

   StoreI4(buf, flags);

   if (flags & FRAMES_FLAG)
      StoreI4(buf, frames);
   if (flags & BYTES_FLAG)
      StoreI4(buf, bytes);

   if (flags & TOC_FLAG)
      memcpy(buf, toc, 100), buf+=100;

   if (flags & VBR_SCALE_FLAG)
      StoreI4(buf, vbr_scale);
}

void XHEADDATA::Recalculate(size_t _frames, size_t _bytes, size_t* frameloc, size_t offset)
{  frames = _frames;
   bytes = _bytes;
   flags = FRAMES_FLAG | BYTES_FLAG | TOC_FLAG;
   // recreate TOC
   offset -= frameloc[0]; // gives an overflow, this is intended
   for (int i = 0; i < 100; ++i)
      toc[i] = (long long)(frameloc[(i * frames + 50) /100] + offset) * 256 / bytes;
   #ifdef DEBUG
   Print();
   #endif
}

size_t XHEADDATA::Length() const
{  size_t len = 4;

   if (flags & FRAMES_FLAG)
      len += 4;
   if (flags & BYTES_FLAG)
      len += 4;
   if (flags & TOC_FLAG)
      len += 100;
   if (flags & VBR_SCALE_FLAG)
      len += 4;

   return len;
}



// control vars (command line args) *****************************************
static const char*   fname = NULL;  // input file name (required)
static const char*   ofname = NULL; // output file name (optional)
static int           verbose = 0;   // verbose screen output
static bool          quiet = false; // no screen output
static int           strip = 0;     // strip level (/s)
static bool          wxing = false; // rewrite Xing header
static bool          cut = false;   // file cutting enable flag
static int           gainchange = 0;// gain change by integer
static int           z0 = 0;        // make 1st z0 frames silent
static int           startframe = -1;// start time for cut/copy
static int           endframe = -1; // end time for cut/copy
static double        starttime = -1;// start time for cut/copy
static double        endtime = -1;  // end time for cut/copy
static bool          fastmode = false;// Do not analyze frame content

// internal vars
static FILE* fh;                    // input file handle
static union
{  MPEG_Frame_Header hdr;
   unsigned char     framebuf[4*65536];
};
static MPEG_Frame_Header hdr_OK;
static int           bitrate;
static int           samprate;
static size_t        framelen;
static size_t        filelen;
static long          frame = 0;
static int           bitrates[16] = {0};
static int           smodes[4] = {0};
static long          gainnum = 0;
static int           gainmin[2] = {256, 256};
static int           gainmax[2] = {-1, -1};
static unsigned long gainavg[2] = {0};
static vector<size_t> frameloc;
struct main_data_info
{  size_t            size;          // main_data_size
   size_t            end;           // main_data_end
   main_data_info(size_t s, size_t e) : size(s), end(e) {}
};
static vector<main_data_info> datainfo; // main_data_end (Layer 3 only)
struct cache_info
{  size_t            Pos;
   ref_array<unsigned char> Data;
   cache_info(size_t pos, const unsigned char* data, size_t len) : Pos(pos), Data(len, data) {
      #ifdef DEBUG
      if (verbose > 1)
         printf("Cache: %lx, %p, %lu\n", pos, data, len);
      #endif
   }
};
static list<cache_info> wcache; // write cache
static long          strip1st = -1;
static long          striplast = -1;
static auto_ptr<XHEADDATA> xing;
static size_t        ID3v2Tag = 0;   // length of ID3v2 data at the beginning of the file if any
static bool          ID3tag = false; // has ID3v1.x tag


static void frameerror(EventClass cls, const char* fmt, ...)
{  fprintf(stderr, "%s at frame %lu (@ %lx): ", Class2String(cls), frame+1, frameloc.size() ? frameloc.back() : 0);
   va_list va;
   va_start(va, fmt);
   vfprintf(stderr, fmt, va);
   va_end(va);
   fputc('\n', stderr);
}


// bitstream helpers ********************************************************
class bitstream
{private:
   unsigned char*    sibuf;
   unsigned int      bit;
   static const unsigned int bitmask[17];
 public:
   bitstream(unsigned char* sibuf, unsigned int bit = 0) : sibuf(sibuf), bit(bit) {}
   unsigned int      get(unsigned int bits);
   unsigned int      peek(unsigned int bits) const;
   void              set(unsigned int bits, unsigned int value);
   unsigned char*    begin() const
                     {  return sibuf;
                     }
   bitstream&        operator+=(unsigned int bits)
                     {  sibuf += bits >> 3;
                        bit = bit+bits & 7;
                        return *this;
                     }
   bitstream         operator+(unsigned int bits) const
                     {  return bitstream(sibuf + (bits>>3), bit+bits & 7);
                     }
};

const unsigned int bitstream::bitmask[17] = {0, 1, 3, 7, 0xf, 0x1f, 0x3f, 0x7f, 0xff, 0x1ff, 0x3ff, 0x7ff, 0xfff, 0x1fff, 0x3fff, 0x7fff, 0xffff};

unsigned int bitstream::get(unsigned int bits)
{  assert(bits <= 16);
   bit += bits;
   unsigned int rval = bits & ~7
    ? ((sibuf[0]<<8 | sibuf[1])<<8 | sibuf[2]) >> 24-bit  // > 8 bit
    : (sibuf[0]<<8 | sibuf[1]) >> 16-bit; // <= 8 bit
   sibuf += bit>>3;
   bit &= 7;
   return rval & bitmask[bits];
}

unsigned int bitstream::peek(unsigned int bits) const
{  assert(bits <= 16);
   return ( bits & ~7
    ? ((sibuf[0]<<8 | sibuf[1])<<8 | sibuf[2]) >> 24-bit-bits  // > 8 bit
    : (sibuf[0]<<8 | sibuf[1]) >> 16-bit-bits ) & bitmask[bits]; // <= 8 bit
}

void bitstream::set(unsigned int bits, unsigned int value)
{  assert(bits <= 8);
   unsigned int rval = (sibuf[0] << 8) | sibuf[1];
   bit += bits;
   unsigned int shift = 16-bit;
   rval = (rval & ~(bitmask[bits] << shift)) | value << shift;
   sibuf[1] = (unsigned char)rval; // write back
   sibuf[0] = (unsigned char)(rval >> 8);
   sibuf += bit>>3;
   bit &= 7;
}



// MP3 sideinfo funcs *******************************************************
static unsigned int handlech(bitstream b, int ch)
{  int val = b.peek(8);
   if (gainmin[ch] > val)
      gainmin[ch] = val;
   if (gainmax[ch] < val)
      gainmax[ch] = val;
   gainavg[ch] += val;
   gainnum += !ch;
   //printf("%i%3i ", ch, val);
   if (!val || !(gainchange|z0)) // if val was 0 it is a good idea to leave val=0
      return val;

   if (z0)
      val -= 64;
   if (gainchange)
      val += gainchange;
   if (val < 0)
   {  frameerror(Warning, "gain underflow (%i).", val);
      val = 0;
   } else if (val > 255)
   {  val = 255;
      frameerror(Error, "gain overflow (%i).", val);
   }
   b.set(8, val);
   return val;
}

static void check_block_type(const bitstream& b)
{  register unsigned int raw = b.peek(3);
   if (raw == 4) // block_type == 0
      frameerror(Error, "block_type == 0 (%u)", raw);
}

class reservoir
{private:
   unsigned char  Buffer[2*944+511];// Bit reservoir, max mp3 frame data len + max bit reservoir
   size_t         Fill;          // Fllstand
   size_t         End;           // Ende des aktuellen Frames
   // End <= Fill <= sizeof Buffer
 public:
                  reservoir();
   void           addframe(unsigned char* data, size_t len, size_t main_data_end);
   void           reqframe(size_t bytes);
   const unsigned char* current() const
                  {  return Buffer;
                  }
   size_t         len() const
                  {  return End;
                  }
};

reservoir::reservoir()
 : Fill(0)
 , End(0)
{}

void reservoir::addframe(unsigned char* data, size_t len, size_t main_data_end)
{
   #ifdef DEBUG
   if (verbose > 1)
      fprintf(stderr, "%p{F=%lu, E=%lu}::addframe(%p, l=%lu, m=%lu)\n", this, Fill, End, data, len, main_data_end);
   #endif
   // 1st: remove old data, keep main_data_end bytes
   End = Fill - End;
   if (main_data_end > End)
      frameerror(Error, "main_data_end larger than the bit reservoir (%lu, main_data_end %lu)", End, main_data_end);
   #ifdef DEBUG
    else if (verbose > 1 && main_data_end < End)
      fprintf(stderr, "Anciliary data: %lu\n", End-main_data_end);
   #endif
   if (main_data_end > Fill)
   {  Fill = 0;
   } else
   {  memmove(Buffer, Buffer+Fill-main_data_end, main_data_end);
      Fill = main_data_end;
   }
   // add data to reservoir
   assert(Fill+len < sizeof Buffer);
   memcpy(Buffer+Fill, data, len);
   Fill += len;
}

void reservoir::reqframe(size_t bytes)
{
   #ifdef DEBUG
   if (verbose > 1)
      fprintf(stderr, "%p{F=%lu, E=%lu}::reqframe(m=%lu)\n", this, Fill, End, bytes);
   #endif
   // store new end
   if (bytes > Fill)
   {  frameerror(Error, "bitreservoir insufficient (%lu, requested %lu)", Fill, bytes);
      End = Fill;
   } else
      End = bytes;
}


static int layerIIIframe()
{  if (hdr.Version < 2)
      return INT_MAX;  // currently only V1.0 and V2.0 supported
   // catch sideinfo
   int sitype = (hdr.Version & 1) | ((hdr.Mode != 3) << 1);
   unsigned int silen = iSideInfoLen[sitype];
   //printf("x1 ");
   if (4 + 2*hdr.HasCRC() + silen > framelen)
      return 0;   // inclomlete frame without sideinfo -> bad frame, no energy
   bitstream si(framebuf +4 + 2*hdr.HasCRC());
   // handle bit reservoir
   static reservoir res;
   unsigned int mde = si.peek(8 + (hdr.Version != 2));
   unsigned int mdlen = (framebuf+framelen) - (si.begin()+silen);
   //printf("x2 %i, %i, %i ", silen, mdlen, mde);
   res.addframe(si.begin()+silen, mdlen, mde);
   //printf("x2a");
   datainfo.push_back(main_data_info(res.len(), mde)); // remember main_data length and end
   // calculate used length
   //printf("x3 ");

   // ignore zero frames (i.e. VBR info's)
   static const char zeros[32] = "";
   if (memcmp(si.begin(), zeros, silen) == 0)
      return 0;

   unsigned int maxgain;
   unsigned int mdused; // used bits in main_data
   //printf("-%i-", sitype);
   switch (sitype)
   {default:// mono, lsf
      mdused = (si+9).peek(12);
      maxgain = handlech(si+(9+21), 0);
      check_block_type(si+(9+38));
      break;
    case 1: // mono
      mdused = (si+18).peek(12) + (si+(18+59)).peek(12);
      maxgain = max(handlech(si+(18+21), 0), handlech(si+(18+59+21), 0));
      check_block_type(si+(18+33));
      check_block_type(si+(18+59+33));
      break;
    case 2: // stereo, lsf
      mdused = (si+10).peek(12) + (si+(10+63)).peek(12);
      maxgain = max(handlech(si+(10+21), 0), handlech(si+(10+63+21), 1));
      check_block_type(si+(10+38));
      check_block_type(si+(10+63+38));
      break;
    case 3: // stereo
      //printf("%u, %u, %u, %u\n", (si+20).peek(12), (si+(20+59)).peek(12), (si+(20+2*59)).peek(12), (si+(20+3*59)).peek(12));
      mdused = (si+20).peek(12) + (si+(20+59)).peek(12) + (si+(20+2*59)).peek(12) + (si+(20+3*59)).peek(12);
      maxgain = max( max(handlech(si+(20+21), 0), handlech(si+(20+59+21), 1)),
                     max(handlech(si+(20+2*59+21), 0), handlech(si+(20+3*59+21), 1)) );
      check_block_type(si+(20+33));
      check_block_type(si+(20+59+33));
      check_block_type(si+(20+2*59+33));
      check_block_type(si+(20+3*59+33));
   }
   //printf("x4 ");
   // check main data size
   res.reqframe((mdused+7)>>3);
   return maxgain;
}

#include "l2tables.h"

static void handlescalefac(bitstream& frm, int n)
{  if (gainchange)
      while (n--)
      {  int val = 63-frm.peek(6);
         if (gainmin[0] > val)
            gainmin[0] = val;
         if (gainmax[0] < val)
            gainmax[0] = val;
         gainavg[0] += val;
         ++gainnum;
         if (z0)
            val -= 32;
         val += gainchange;
         if (val < 0)
         {  frameerror(Warning, "scalefactor underflow (%i).", val);
            val = 0;
         } else if (val > 63)
         {  frameerror(Error, "scalefactor overflow (%i).", val);
            val = 63;
         }
         frm.set(6, 63-val); // write back
      }
    else
      while (n--)
      {  int val = 63-frm.get(6);
         if (gainmin[0] > val)
            gainmin[0] = val;
         if (gainmax[0] < val)
            gainmax[0] = val;
         gainavg[0] += val;
         ++gainnum;
      }
}

static int layerIIframe()
{
   static const unsigned char translate [3] [2] [16] =
    { { { 0,2,2,2,2,2,2,0,0,0,1,1,1,1,1,0 } ,
        { 0,2,2,0,0,0,1,1,1,1,1,1,1,1,1,0 } } ,
      { { 0,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0 } ,
        { 0,2,2,0,0,0,0,0,0,0,0,0,0,0,0,0 } } ,
      { { 0,3,3,3,3,3,3,0,0,0,1,1,1,1,1,0 } ,
        { 0,3,3,0,0,0,1,1,1,1,1,1,1,1,1,0 } } };

   static const struct al_table2 *tables[5] =
        { alloc_0, alloc_1, alloc_2, alloc_3 , alloc_4 };
   static int sblims[5] = { 27 , 30 , 8, 12 , 30 };

   // audio data segment
   bitstream frm(framebuf + (hdr.HasCRC() ? 4+2 : 4));    // skip CRC

   // fetch allocation table and subband limits
   int table = hdr.Version != 3 // Version > 1.0 => low frequency
    ? 4
    : translate[hdr.SamplingRate][hdr.Mode == 3][hdr.BitRate];

   int sblim = sblims[table];
   const al_table2* alloc = tables[table];
   int jsbound;
   switch (hdr.Mode)
   {case 3: // mono
      jsbound = 0;
      break;
    case 1: // intensity stereo
      jsbound = (hdr.ModeExtension+1) << 2;
      break;
    default:
      jsbound = sblim;
   }
   //printf("at: %i, sblim: %i, jsb: %i\n", table, sblim, jsbound);
   // get the total number of scale factor selection informations
   int nScfsi = 0;
   int step;
   for (int i = jsbound; i--; alloc += 1<<step)
   {  step = alloc->bits;
      nScfsi += (frm.get(step) != 0) + (frm.get(step) != 0);
   }
   if (jsbound < sblim) // avoid crash on nonsens data
      for (int i = sblim-jsbound; i--; alloc += 1<<step)
         nScfsi += frm.get(step = alloc->bits) != 0;

   // get the total number of scalefactors
   int nScaleFac = 0;
   static const char mapScfsi[4] = {3,2,1,2}; // map scfsi to the number of scalefactors
   while (nScfsi--)
      nScaleFac += mapScfsi[frm.get(2)];

   // scale!
   handlescalefac(frm, nScaleFac);

   return nScaleFac;
}

static int layerIframe()
{  if (hdr.Version != 3)
      return INT_MAX;  // currently only V1.0 supported
   // audio data segment
   bitstream frm(framebuf + (hdr.HasCRC() ? 4+2 : 4));    // skip CRC
   // determine the total number of subband allocation coefficients
   int nSubBandChannels = 32; // subbands per channel
   switch (hdr.Mode)
   {case 2: // dual channel
    case 0: // stereo
      nSubBandChannels <<= 1; // twice as much
      break;
    case 1: // intensity stereo
      nSubBandChannels += (hdr.ModeExtension+1) << 2; // number of stereo subbands = 4 * (ModeExtension+1)
      break;
   }
   // fetch the number of scalefactors := number of Subchannels with allocation != 0
   int nScaleFac = 0;
   while (nSubBandChannels--)
      nScaleFac += frm.get(4) != 0;
   // scale!
   handlescalefac(frm, nScaleFac);

   return nScaleFac;
}

static void commitframe(const unsigned char* frm, const unsigned char* bak, size_t len)
{  if (!gainchange && !z0)
      return;  // nothing to do
   const unsigned char* p1 = frm;
   const unsigned char* p2 = p1 + len;
   const unsigned char* b = bak;
   for (;;) // find first different
   {  if (p1 == p2)
         return;
      if (*p1 != *b)
         break;
      ++p1; ++b;
   }
   b = bak + len;
   while (*--p2 == *--b); // find last different
   wcache.push_back(cache_info(p1 - frm + (frameloc.size() ? frameloc.back() : 0), p1, p2 - p1 +1));
}

static void handleframe()
{  // xing check
   if (!frame && !xing.get())
      xing = XHEADDATA::Get(framebuf);

   // save frame data
   static unsigned char framebak[2048]; // hopefully sufficient
   memcpy(framebak, framebuf, framelen);

   int maxgain = INT_MAX;
   switch (hdr.GetLayer())
   {case 3:
      maxgain = layerIIIframe(); break;
    case 2:
      maxgain = layerIIframe(); break;
    case 1:
      maxgain = layerIframe(); break;
   }

   // record frame changes
   commitframe(framebuf, framebak, framelen);

   if (z0)
      --z0;

   // analyze for strip silence
   if (strip)
   {  int sum = 0;
      if (strip < 0 || (maxgain < 1 && strip1st == -1 && hdr.Layer == 1))
      {  // scan frame (because of bit layer III reservoir)
         unsigned char* cp = framebuf + (hdr.HasCRC() ? 4+2 : 4);
         unsigned char last = 0;
         int i = framebuf + framelen - cp;
         while (i--)
         {  sum += *cp != last;
            last = *cp++;
         }
      }  // store result
      if (sum > abs(strip) || (strip > 0 && maxgain >= 1))
      {  striplast = frame +1;
         if (strip1st == -1)
            strip1st = frame;
   }  }
}

static void printframeinfo(const MPEG_Frame_Header hdr)
{  printf("%-7li%-6u%-5.4s%-7.4s%-5.4s%3i   %5i  %-5.4s%-5i%-6.2s%-10.8s%-6.5s%.5s\n",
    frame, framelen, sVersion[hdr.Version], sLayer[hdr.Layer], sYesNo[!hdr.CRC],
    bitrate, max(iSampleRate[hdr.Version][hdr.SamplingRate], 0),
    sYesNo[hdr.Padding], hdr.Private, sMode[hdr.Mode],
    hdr.Mode == 1 ? sExtMode[hdr.Layer <= 1][hdr.ModeExtension] : "",
    sCopyright[hdr.Copyright], sEmphasis[hdr.Emphasis]);
}

static void filedatamove(FILE* fi, FILE* fo, size_t len, size_t offset)
{
   #ifdef DEBUG
   printf("filedatamove: %lu, %lu\n", len, offset);
   #endif
   // skip old patches
   for ( list<cache_info>::iterator i = wcache.begin();
      i != wcache.end() && i->Pos+i->Data.size() < offset;
      i = wcache.erase(i) ); // next
   // copy data
   while (len)
   {  // determine block size
      const size_t l = min(sizeof framebuf, len);
      // readsource data
      if (fread(framebuf, l, 1, fi) != 1)
         throw Event(27, "FATAL: Read %lu failed (%i).", l, errno);
      // look for patches
      for ( list<cache_info>::iterator i = wcache.begin();
         i != wcache.end() && i->Pos < offset+l;
         i = wcache.erase(i) ) // next
      {  long diff = i->Pos - offset;
         if (diff < 0)
            memcpy(framebuf, i->Data.get() - diff, min((size_t)(i->Data.size() + diff), l)); // starts before
          else
            memcpy(framebuf + diff, i->Data.get(), min(i->Data.size(), size_t(l - diff))); // starts after
         if (diff + i->Data.size() > offset+l)
            break; // keep last entry
      }
      // write result
      if (fwrite(framebuf, l, 1, fo) != 1)
         throw Event(27, "FATAL: Write %lu failed (%i).", l, errno);
      len -= l;
      offset += l;
}  }

static void applypatches(FILE* fo, size_t len)
{
   #ifdef DEBUG
   printf("applypatches: %lu\n", len);
   #endif
   for ( list<cache_info>::iterator i = wcache.begin();
      i != wcache.end() && i->Pos < len;
      i = wcache.erase(i) ) // next
   {  if (fseek(fo, i->Pos, SEEK_SET))
         throw Event(27, "FATAL: Seek to %lu failed.", i->Pos);
      #ifdef DEBUG
      if (verbose > 1)
         printf("Patch: @%lx, %lu\n", i->Pos, i->Data.size());
      #endif
      if (fwrite(i->Data.get(), min(i->Data.size(), len - i->Pos), 1, fo) != 1)
         throw Event(27, "FATAL: Write failed (%i).", errno);
   }
   // known bug: this function will purge partial patches, too.
}


// command line parsing funcs ***********************************************
static int intoption(const char* arg, const char* context)
{  size_t n = 0;
   int r;
   sscanf(arg, "%i%n", &r, &n);
   if (n != strlen(arg))
      throw Event(42, "Integer constant expected at %s.", context);
   return r;
}

static double timeoption(const char* arg, const char* context)
{  double t = 0;
   size_t n;
   for (;;)
   {  const char* pc = strchr(arg, ':');
      if (pc == NULL)
         break;
      unsigned u;
      if (sscanf(arg, "%u%n", &u, &n) != 1 || n != (size_t)(pc-arg))
         throw Event(42, "Time constant %.*s must be a number at %s.", pc-arg, arg, context);
      t = 60*t + u;
      arg = pc+1;
   }
   double d;
   if (sscanf(arg, "%lf%n", &d, &n) != 1 || n != strlen(arg))
      throw Event(42, "Time constant %s must be a number at %s.", arg, context);
   return 60*t + d;
}

static int parseoption(const char* cp)
{  switch (tolower(cp[0]))
   {case 'v':
      verbose = 1;
      return 0;
    case '_':
      verbose = 2;
      return 0;
    case 'q':
      quiet = true;
      return 0;
    case 's':
      strip = cp[1] == 0
       ? -9
       : intoption(cp+1, cp);
      return 0;
    case 'c':
      cut = true;
      return 0;
    case 'f':
      switch (tolower(cp[1]))
      {case 's':
         startframe = intoption(cp+2, cp);
         return 0;
       case 'e':
         endframe = intoption(cp+2, cp);
         return 0;
      };
      break; // -> Error
    case 't':
      switch (tolower(cp[1]))
      {case 's':
         starttime = timeoption(cp+2, cp);
         return 0;
       case 'e':
         endtime = timeoption(cp+2, cp);
         return 0;
      };
      break; // -> Error
    case 'l':
      gainchange = intoption(cp+1, cp);
      return 0;
    case 'z':
      z0 = cp[1] == 0
       ? 1
       : intoption(cp+1, cp);
      return 0;
    case 'n':
      fastmode = true;
      return 0;
    default:;
   }
   throw Event(44, "Illegal option %s.", cp);
}


int _cdecl main(int argc, char** argv)
{
   #ifdef DEBUG
   {  // verify wether min/max are side effect free
      int i = 1;
      int j = 2;
      int k = min(++i, ++j);
      int l = min(++j, ++i);
      assert(i == 3 && j == 4 && k == 2 && l == 3);
   }
   #endif

   // ***** parse command line ...
   if (argc < 2)
   {  fputs("Usage: FrameAnalyzer <filename> [<outfilename>] [<options>]\n"
            "options: /l#   change gain by 1.5*# dB\n"
            "         /z[#] make the first (#) frame(s) silent\n"
            "         /s[#] strip empty frames (threshold #)\n"
            "         /ts#  start at # seconds (# my be [h:]m:s[.f])\n"
            "         /te#  end at # seconds (# my be [h:]m:s[.f])\n"
            "         /fs#  start at frame #\n"
            "         /fe#  end at frame #\n"
            "         /n    fast mode, do not analyze frame content\n"
            "         /v    print info for every frame\n"
       , stderr);
      return 10;
   }
   try
   {  for (int i = 1; i < argc; ++i)
      {  char* cp = argv[i];
         if ((cp[0] == '/' || cp[0] == '-') && cp[1] != 0) // dont't treat '-' as option
         {  int r = parseoption(cp+1);
            if (r)
               return r;
         } else if (ofname)
         {  throw Event(49, "Syntax error in command line.");
         } else if (fname)
         {  ofname = cp;
            cut = true;
         } else
         {  fname = cp;
      }  }

      // ***** check parameter constraints ...
      if (!fname)
         throw Event(40, "Filename required.");
      if (fname[0] == '-' && fname[1] == 0)
         throw Event(30, "Soure can't be stdin or any other transient stream.");
      if ((startframe >= 0 || endframe >= 0 || starttime >= 0 || endtime >= 0) && !cut)
         fputs("Warning: you must enable the frame and/or time cut options (/f.. or /t..) explicitly by /c when operating in-place.\n"
               "         The file will not be striped.\n", stderr);
      if (fastmode && (z0 | strip | gainchange))
         throw Event(34, "The options /l, /z and /s cannot  operate in fast mode.");

      // ***** source analysis ...

      // open file
      int fhi = _sopen(fname, O_RDONLY|O_BINARY, SH_DENYNO, 0);
      if (fhi == -1)
         throw Event(20, "Error opening %s.", fname);
      fh = fdopen(fhi, "rb");
      setvbuf(fh, NULL, _IOFBF, 262144); // increase file buffer
      fprintf(stderr, "Scanning %s...\r", fname);

      // read V1.x tag if any
      unsigned char tag[128];
      if (fseek(fh, -128, SEEK_END))
         throw Event(27, "Seek to V1.x tag failed.");
      if (fread(&tag, 128, 1, fh) != 1)
         throw Event(27, "Read V1.x tag failed.");
      filelen = ftell(fh);
      rewind(fh);
      if (memcmp(tag, "TAG", 3) == 0)
      {  ID3tag = true;
         if (!quiet)
            printf("ID3 Tag:\n"
                   " Title:  %.30s\n"
                   " Artist: %.30s\n"
                   " Album:  %.30s\n"
                   " Year:   %.4s\n"
                   " Comment:%.30s\n\n", tag+3, tag+33, tag+63, tag+93, tag+97);
         filelen -= 128;
      }
      // skip V2.x tag if any
      unsigned char tag2[10];
      if (fread(&tag2, 10, 1, fh) == 1)
      {  if (memcmp(tag2, "ID3", 3) == 0 && tag2[3] != 0xff && tag2[4] != 0xff && (tag2[6]|tag2[7]|tag2[8]|tag2[9]) < 0x80)
            ID3v2Tag = 10 + tag2[9] + ((tag2[8] + (tag2[7] + (tag2[6] << 7) << 7)) << 7);
         if (fseek(fh, ID3v2Tag, SEEK_SET))
            throw Event(27, "Skip of V2.x tag failed.");
      }

      // scan file
      if (!quiet && verbose)
         puts("Frame  len.  ver. layer  CRC  kbps  sampl. pad. pri. mode  IS  MS-S  (C)   emph.");
      while((size_t)ftell(fh) < filelen)
      {  // read MPEG Header
         if (fread(&hdr, 4, 1, fh) != 1)
            break;
         // check header
         if ((hdr.FrameSync1 & hdr.FrameSync2) != -1)
         {  frameerror(Error, "illegal frame sync 0x%02x%x (@ %lx)", hdr.FrameSync1&0xFF, hdr.FrameSync2&7, ftell(fh)-4);
            fseek(fh, -3, SEEK_CUR);
            // try resync
          resync:
            fscanf(fh, "%*[^\377]\377");
          resync2:
            int c = getc(fh);
            if (c == EOF)
               break;
            if (c == 0xFF) // gcc runtime fix
               goto resync2;
            if (c < 0xE0)
               goto resync;
            c = getc(fh);
            if (c == EOF)
               break;
            if (c == 0xFF)
               goto resync2;
            // found possible sync
            fseek(fh, -3, SEEK_CUR);
            fprintf(stderr, "Resync at %lX\n", ftell(fh));
            continue;
         }
         bitrate = hdr.GetBitRate();
         switch (bitrate)
         {case -1:
            fputs("bitrate bad (0xF)\n", stderr);
          case -2:
            bitrate = 0;
          default:;
         }
         samprate = hdr.GetSamplingRate();
         switch (samprate)
         {case -1:
            fputs("sampling rate bad (0xF)\n", stderr);
          case -2:
            samprate = 0;
          default:;
         }
         framelen = hdr.GetFrameLen(bitrate);

         // print header
         if (verbose)
            printframeinfo(hdr);

         if (framelen > 4)
         {  // frame seems to be valid
            hdr_OK = hdr;
            size_t rem = ftell(fh) -4;
            #ifdef DEBUG
            if (verbose > 1)
               printf("Frame: %lx, %08x\n", rem, *(int*)&hdr_OK);
            #endif
            frameloc.push_back(rem);
            rem = filelen - rem;
            if (framelen > rem)
            {  memset(framebuf +4, 0, framelen -4);
               framelen = rem;
               fputs("incomplete frame\n", stderr);
            }
            // statistic
            ++bitrates[hdr.BitRate];
            ++smodes[hdr.ModeExtension];

            if (fastmode)
            {  // skip frame content
               if (fseek(fh, framelen -4, SEEK_CUR) == -1)
               {  fputs("Failed to advance to next frame.\n", stderr);
                  break;
               }
            } else
            {  // read frame
               if (fread(framebuf +4, framelen -4, 1, fh) != 1)
               {  fputs("Failed to read frame data.\n", stderr);
                  break;
               }
               // analyze sideinfo
               handleframe();
            }

            // next frame
            ++frame;
         }
      }

      if (!frame)
         throw Event(10, "Warning: File %s contains no MPEG Frames.", fname);

      // restore last valid values
      const int* brvect = hdr_OK.GetBitRates();
      bitrate = hdr_OK.GetBitRate(brvect);
      samprate = hdr_OK.GetSamplingRate();
      framelen = hdr_OK.GetFrameLen(bitrate);
      // last frame end
      frameloc.push_back(min(frameloc.back() + framelen, filelen));
      // print last header
      if (!verbose && !quiet)
      {  puts("Frame  len.  ver. layer  CRC  kbps  sampl. pad. pri. mode  IS  MS-S  (C)   emph");
         printframeinfo(hdr_OK);
      }
      // statistics
      if (!quiet)
      {  unsigned long long hundrets = ((unsigned long long)hdr_OK.GetFrameSamples() * frame * 4) / (hdr_OK.GetSamplingRate() / 25);
         printf("total: %2lu:%02lu:%05.2f [h:mm:ss.ss]\n"
          , (unsigned long)(hundrets / 360000LU), (unsigned long)(hundrets / 6000LU % 60LU), hundrets % 6000U / 100.);
         if (hdr_OK.Mode == 1)
         {  puts("\nJS-Mode  frames     %");
            for (int i = 0; i < 4; ++i)
               if (smodes[i])
                  printf("%-8.8s %6i %5.1f\n"
                   , sSExtMode[hdr_OK.Layer == 1][i], smodes[i], 100.*smodes[i]/frame);
         }
         if (bitrates[hdr_OK.BitRate] != (int)frame)
         {  char buf1[80] = "Rate", buf2[80] = "Frms", buf3[80] = "%   ";
            int n1 = 4, n2 = 4, n3 = 4;
            for (int i = 1; i < 15; ++i)
               if (bitrates[i])
               {  n1 += sprintf(buf1 + n1, " %4u", brvect[i]);
                  n2 += (bitrates[i] >= 10000)
                   ? sprintf(buf2 + n2, " %3uk", (bitrates[i]+500)/1000)
                   : sprintf(buf2 + n2, " %4u", bitrates[i]);
                  n3 += sprintf(buf3 + n3, "%5.1f", 100.*bitrates[i]/frame);
               }
            printf("\n%s\n%s\n%s\n", buf1, buf2, buf3);
            clearerr(fh);
            printf("average: %5.1f [kbps]\n"
             , (frameloc.back() - frameloc.front()) * .8 / hundrets);
         }
         if (gainnum)
         {  printf("\nGain  min.  max.  average\n"
                     "left  %-4i  %-4i  %.1f\n"
             , gainmin[0], gainmax[0], (double)gainavg[0]/gainnum);
            if (hdr_OK.Has2Channels())
               // stereo
               printf("right %-4i  %-4i  %.1f\n"
                , gainmin[1], gainmax[1], (double)gainavg[1]/gainnum);
      }  }

      if (!frame)
         throw Event(24, "Stop: File seems to contain not even one MPEG frame.");

      // ***** writing modifications ...

      // prepare strip/cut points if any
      if (cut)
      {  double frametime = (double)hdr_OK.GetFrameSamples() / hdr_OK.GetSamplingRate(); // time [s] per frame
         #ifdef DEBUG
         printf("FT: %f s/et: %f/%f s/ef: %u/%u\n", frametime, starttime, endtime, startframe, endframe);
         #endif
         if (starttime >= 0)
         {  long f = (long)(starttime/frametime +.5);
            if (f > frame)
               throw Event(24, "Start time is outside the file.");
            if (f > strip1st) // store
               strip1st = f;
         }
         if (startframe >= 0)
         {  if (startframe > frame)
               throw Event(24, "Start frame is outside the file.");
            if (startframe > strip1st) // store
               strip1st = startframe;
         }
         if (endtime >= 0)
         {  long f = (long)(endtime/frametime +.5);
            if (f > frame)
               throw Event(24, "End time is outside the file.");
            if (striplast == -1 || f < striplast) // store
               striplast = f;
         }
         if (endframe >= 0)
         {  if (endframe > frame)
               throw Event(24, "End frame is outside the file.");
            if (striplast == -1 || endframe < striplast) // store
               striplast = endframe;
      }  }
      bool stripflag = strip1st != -1 || striplast != -1;
      if (strip1st == -1) // defaults -> file begin/end
         strip1st = 0;
      if (striplast == -1)
         striplast = frame;
      // *** now [strip1st, striplast) is the destination range
      if (strip1st >= striplast)
         throw Event(14, "Error: desired file range is empty [%i,%i).", strip1st, striplast);
      // strip really required ?
      stripflag &= strip1st > (xing.get() != NULL) || striplast != frame || frameloc.front() > ID3v2Tag;
      #ifdef DEBUG
      printf("Strip: %li, %li, %i\n", strip1st, striplast, stripflag);
      #endif

      // destination file :=
      //  ID3v2 header (if any)
      //  Xing header (if any)
      //  frames[strip1st, striplast)
      //  ID3V1 tag (if any)

      size_t filestart = ID3v2Tag; // starting point of MPEG data
      if (xing.get())
      {  if (strip1st == 0) // whatever happens, do not duplicate the Xing header
            strip1st = 1;
         filestart = frameloc[1];
      }
      size_t datalen = frameloc[striplast] - frameloc[strip1st];

      // recalc Xing header if any
      if (xing.get() && (wxing || stripflag))
      {  // dirty hack: patch frameloc[strip1st-1] to match the size of the Xing header
         frameloc[strip1st-1] = frameloc[strip1st] - (frameloc[1] - frameloc[0]);
         // total data length
         xing->Recalculate(striplast - strip1st +1,   // + Xing header
          filestart + datalen + 128*ID3tag,           // total file length
          &frameloc[0] + strip1st -1,                 // Xing header => -1
          ID3v2Tag);
         // store in write cache
         size_t off = xing->Offset();
         xing->Put(framebuf);
         wcache.push_front(cache_info(filestart + off, framebuf + off, xing->Length()));
      }

      // write required?
      if (stripflag || wcache.size() || ofname)
      {  // open output file (maybe the same than input)
         int fiho = -1;
         FILE* fho;
         if (ofname != NULL)
         {  if (ofname[0] == '-' && ofname[1] == 0)   // stdout ?
               fho = stdout;
             else
            {  fiho = _sopen(ofname, O_RDWR|O_BINARY|O_CREAT, SH_DENYWR, S_IREAD|S_IWRITE);
               if (fiho == -1)
                  throw Event(20, "Error creating %s (%i).", ofname, errno);
               fho = fdopen(fiho, "r+b");
            }
         } else // in-place mode
         {  fiho = _sopen(fname, O_RDWR|O_BINARY, SH_DENYWR, S_IREAD|S_IWRITE);
            if (fiho == -1)
               throw Event(20, "Error opening %s for writing (%i).", fname, errno);
            fho = fdopen(fiho, "r+b");
         }
         setvbuf(fho, NULL, _IOFBF, 262144); // increase file buffer

         fprintf(stderr, "Writing %s, frame %lu to %lu\r", ofname ? ofname : fname, strip1st, striplast);

         // write headers
         if (ofname)
         {  // not in place
            rewind(fh);
            filedatamove(fh, fho, filestart, 0);
         } else
         {  // otherwise apply header patches only (e.g. Xing)
            applypatches(fho, filestart);
            if (fseek(fho, filestart, SEEK_SET))
               throw Event(27, "FATAL: Seek to header end failed (%i).", errno);
         }
         // now fho points just behind the headers

         // process main data
         if (ofname || frameloc[strip1st] > filestart)
         {  // fix bit reservoir
            /*if (hdr_OK.GetLayer() == 3) // layer III & no /z option
            {  size_t rsize = datainfo[strip1st].size - datainfo[strip1st].end; // total number of bytes in the past at frame strip1st
               // we have to get room for additional rsize bytes
               fprintf(stderr, "TODO: fix bit reservoir. We need another %lu bytes.\n", rsize);
            }*/

            // MPEG data not in place
            if (fseek(fh, frameloc[strip1st], SEEK_SET))
               throw Event(27, "FATAL: Seek to MPEG data start failed (%i).", errno);
            filedatamove(fh, fho, datalen, frameloc[strip1st]);
         } else
         {  // MPEG data in place
            applypatches(fho, filestart + datalen);
            if (stripflag && ID3tag && fseek(fho, filestart + datalen, SEEK_SET))
               throw Event(27, "FATAL: Seek to MPEG data end failed (%i).", errno);
         }

         // write ID3v1.x tag if any
         if ((ofname || stripflag) && ID3tag && fwrite(&tag, 128, 1, fho) != 1)
            throw Event(27, "FATAL: Write tag failed (%i).", errno);
         // adjust file size
         if (fiho != -1)
            ftruncate(fiho, filestart + datalen + 128*ID3tag); // requires BSD compatible funcs

         fclose(fho);   // fin
      }
      fclose(fh);

   } catch (const Event& err)
   {  fprintf(stderr, "%s\n", err.buf);
      return err.err;
   }

   return 0;
}

